from io import BytesIO

import pandas as pd

from wbportfolio.models import Product, Trade


def parse(import_source):
    df_dict = pd.read_excel(BytesIO(import_source.file.read()), engine="openpyxl")
    df_dict = df_dict.rename(
        columns={
            "Trade Date": "transaction_date",
            "ISIN": "underlying_instrument__isin",
            "Client Side": "client_side",
            "Net Quantity": "shares",
            "Price": "price",
            "Custodian": "custodian",
        }
    )
    df_dict = df_dict.where(pd.notnull(df_dict), None)
    df_dict["transaction_date"] = pd.to_datetime(df_dict["transaction_date"])
    product_isins = set()

    data = list()
    for obj in df_dict.to_dict("records"):
        shares = -1 * obj["shares"]
        isin = obj["underlying_instrument__isin"]
        data.append(
            {
                "underlying_instrument": {"isin": isin, "instrument_type": "product"},
                "portfolio": {"isin": isin, "instrument_type": "product"},
                "transaction_date": obj["transaction_date"].strftime("%Y-%m-%d"),
                "shares": shares,
                "bank": obj["custodian"] if obj["custodian"] else "NA",
                "transaction_subtype": Trade.Type.REDEMPTION if shares < 0 else Trade.Type.SUBSCRIPTION,
                "price": round(obj["price"], 6),
            }
        )
        product_isins.add(isin)

    underlying_instruments = Product.objects.filter(isin__in=product_isins).values_list("id", flat=True)

    return {
        "data": data,
        "history": {
            "underlying_instruments": list(underlying_instruments),
            "transaction_date": df_dict.transaction_date.max().strftime("%Y-%m-%d"),
        },
    }



from django.dispatch import receiver
from wbcore.contrib.currency.factories import CurrencyFXRatesFactory
from wbcore.contrib.directory.factories import CompanyFactory
from wbcore.test.signals import custom_update_kwargs, get_custom_factory
from wbfdm.factories import InstrumentPriceFactory

from wbportfolio.factories import (
    CustomerTradeFactory,
    OrderProposalFactory,
    ProductFactory,
    ProductPortfolioRoleFactory,
)
from wbportfolio.viewsets import (
    AssetPositionInstrumentModelViewSet,
    AssetPositionUnderlyingInstrumentChartViewSet,
    AUMProductChartView,
    ClaimEntryModelViewSet,
    CustodianDistributionInstrumentChartViewSet,
    CustomerDistributionInstrumentChartViewSet,
    DistributionChartViewSet,
    DistributionTableViewSet,
    NominalProductChartView,
    OrderOrderProposalModelViewSet,
    OrderProposalPortfolioModelViewSet,
    PortfolioRoleInstrumentModelViewSet,
    ProductCustomerModelViewSet,
    ProductPerformanceFeesModelViewSet,
    SubscriptionRedemptionInstrumentModelViewSet,
    SubscriptionRedemptionModelViewSet,
    TradeInstrumentModelViewSet,
)

# =================================================================================================================
#                                              CUSTOM FACTORY
# =================================================================================================================


@receiver(get_custom_factory, sender=SubscriptionRedemptionModelViewSet)
@receiver(get_custom_factory, sender=SubscriptionRedemptionInstrumentModelViewSet)
def receive_factory_subscription_redemption_instrument(sender, *args, **kwargs):
    return CustomerTradeFactory


@receiver(get_custom_factory, sender=PortfolioRoleInstrumentModelViewSet)
def receive_factory_product_portfolio_role(sender, *args, **kwargs):
    return ProductPortfolioRoleFactory


# =================================================================================================================
#                                              UPDATE DATA
# =================================================================================================================


# =================================================================================================================
#                                              UPDATE KWARGS
# =================================================================================================================
@receiver(custom_update_kwargs, sender=OrderOrderProposalModelViewSet)
def receive_kwargs_trade_order_proposal_product(sender, *args, **kwargs):
    if obj := kwargs.get("obj_factory"):
        order_proposal = OrderProposalFactory.create()
        obj.order_proposal = order_proposal
        obj.save()
        return {"order_proposal_id": order_proposal.id}
    return {}


@receiver(custom_update_kwargs, sender=OrderProposalPortfolioModelViewSet)
def receive_kwargs_trade_order_proposal_portfolio(sender, *args, **kwargs):
    if obj := kwargs.get("obj_factory"):
        return {"portfolio_id": obj.portfolio.id}
    return {}


@receiver(custom_update_kwargs, sender=TradeInstrumentModelViewSet)
@receiver(custom_update_kwargs, sender=AssetPositionInstrumentModelViewSet)
@receiver(custom_update_kwargs, sender=SubscriptionRedemptionInstrumentModelViewSet)
@receiver(custom_update_kwargs, sender=AssetPositionUnderlyingInstrumentChartViewSet)
def receive_kwargs_instrument(sender, *args, **kwargs):
    if kwargs.get("underlying_instrument_id"):
        return {"instrument_id": kwargs.get("underlying_instrument_id")}
    return {}


@receiver(custom_update_kwargs, sender=ProductCustomerModelViewSet)
def receive_kwargs_product_customer_instrument(sender, *args, **kwargs):
    if obj := kwargs.get("obj_factory"):
        InstrumentPriceFactory.create(instrument=obj, calculated=False)
        obj.update_last_valuation_date()
    return {}


@receiver(custom_update_kwargs, sender=PortfolioRoleInstrumentModelViewSet)
def receive_kwargs_portfolio_role_instrument(sender, *args, **kwargs):
    if obj := kwargs.get("obj_factory"):
        return {"instrument_id": obj.instrument.id}
    return {}


@receiver(custom_update_kwargs, sender=DistributionChartViewSet)
@receiver(custom_update_kwargs, sender=DistributionTableViewSet)
def receive_kwargs_distribution_chart_instrument_id(sender, *args, **kwargs):
    if instrument_id := kwargs.get("underlying_instrument_id"):
        return {"instrument_id": instrument_id}
    return {}


@receiver(custom_update_kwargs, sender=ProductPerformanceFeesModelViewSet)
def receive_kwargs_product_performance_fees(sender, *args, **kwargs):
    CurrencyFXRatesFactory()
    return {}


@receiver(custom_update_kwargs, sender=CustodianDistributionInstrumentChartViewSet)
def receive_kwargs_custodian_distribution_instrument(sender, *args, **kwargs):
    if instrument_id := kwargs.get("underlying_instrument_id"):
        return {"instrument_id": instrument_id}
    return {}


@receiver(custom_update_kwargs, sender=NominalProductChartView)
@receiver(custom_update_kwargs, sender=AUMProductChartView)
def receive_kwargs_aum_product(sender, *args, **kwargs):
    if obj := kwargs.get("obj_factory"):
        product = ProductFactory()
        obj.instrument = product
        obj.save()
        return {"product_id": product.id}
    return {}


@receiver(custom_update_kwargs, sender=ClaimEntryModelViewSet)
def receive_kwargs_claim_entry(sender, *args, **kwargs):
    if claim := kwargs.get("obj_factory"):
        owner = CompanyFactory.create()
        claim.account.owner = owner
        claim.account.save()
        return {"entry_id": owner.id}
    return {}


@receiver(custom_update_kwargs, sender=CustomerDistributionInstrumentChartViewSet)
def receive_kwargs_customer_distribution(sender, *args, **kwargs):
    if trade := kwargs.get("obj_factory"):
        return {"instrument_id": trade.underlying_instrument.id}
    return {}



# Generated by Django 5.0.12 on 2025-02-14 13:10

import django_fsm
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('wbportfolio', '0073_remove_product_price_computation_and_more'),
    ]

    operations = [
        migrations.AlterField(
            model_name='rebalancer',
            name='frequency',
            field=models.CharField(default='RRULE:FREQ=MONTHLY;BYDAY=MO,TU,WE,TH,FR;BYSETPOS=1', help_text='The Evaluation Frequency in RRULE format', max_length=256, verbose_name='Evaluation Frequency'),
        ),
        migrations.AlterField(
            model_name='tradeproposal',
            name='status',
            field=django_fsm.FSMField(choices=[('DRAFT', 'Draft'), ('SUBMIT', 'Submit'), ('APPROVED', 'Approved'), ('DENIED', 'Denied'), ('FAILED', 'Failed')], default='DRAFT', max_length=50, verbose_name='Status'),
        ),
    ]



from wbcore.menus import ItemPermission, MenuItem
from wbcore.permissions.shortcuts import is_internal_user

PRODUCTCUSTOMER_MENUITEM = MenuItem(
    label="Product Lists",
    endpoint="wbportfolio:productcustomer-list",
    permission=ItemPermission(method=lambda request: not is_internal_user(request.user)),
)
PRODUCT_MENUITEM = MenuItem(
    label="Products",
    endpoint="wbportfolio:product-list",
    permission=ItemPermission(
        method=lambda request: is_internal_user(request.user), permissions=["wbportfolio.view_product"]
    ),
)



import pytest
from wbcore.test import GenerateTest, default_config

config = {}
for key, value in default_config.items():
    config[key] = list(
        filter(
            lambda x: x.__module__.startswith("wbportfolio")
            and x.__name__
            not in [
                "AggregatedAssetPositionLiquidityPandasView",
                "OrderOrderProposalModelViewSet",
                "PortfolioSwingPricing",
                "PortfolioCashTarget",
                "PortfolioSwingPricingModelViewSet",
                "AccountReconciliationModelViewSet",
                "AccountReconciliationLineModelViewSet",
                "AccountReconciliation",
                "AccountReconciliationLine",
                "TopDownPortfolioCompositionPandasAPIView",
                "CompositionModelPortfolioPandasView",
                # "ClaimModelViewSet",
                # "ClaimModelSerializer",
            ],
            value,
        )
    )


@pytest.mark.django_db
@GenerateTest(config)
class TestProject:
    pass



# Generated by Django 5.0.14 on 2025-06-17 13:58

import django.db.models.deletion
import django.db.models.expressions
import django.utils.timezone
from decimal import Decimal
from django.db import migrations, models
from django.db.migrations.operations.models import ModelOptionOperation
from django.db.models import Count


def remove_duplicate(apps, schema_editor):
    Trade = apps.get_model("wbportfolio", "Trade")
    for row in Trade.objects.filter(trade_proposal__isnull=False).values("underlying_instrument", "transaction_date", "portfolio").annotate(c=Count("*")).filter(c__gt=1):
        qs = Trade.objects.filter(underlying_instrument=row["underlying_instrument"], transaction_date=row["transaction_date"], portfolio=row["portfolio"])
        for t in qs[1:]:
            t.delete()

class SetModelBasesOptionOperation(ModelOptionOperation):
    """
    A migration operation that updates the bases of a model.
    This can be used to separate a model from its parent. Specifically
    when multi-table inheritance is used.
    """
    def __init__(self, name, bases):
        super().__init__(name)
        self.bases = bases

    def deconstruct(self):
        return (self.__class__.__qualname__, [], {"bases": self.bases})

    def state_forwards(self, app_label, state):
        model_state = state.models[app_label, self.name_lower]
        model_state.bases = self.bases
        state.reload_model(app_label, self.name_lower, delay=True)

    def database_forwards(self, app_label, schema_editor, from_state, to_state):
        pass

    def database_backwards(self, app_label, schema_editor, from_state, to_state):
        pass

    def describe(self):
        return "Update bases of the model %s" % self.name

    @property
    def migration_name_fragment(self):
        return "set_%s_bases" % self.name_lower

class Migration(migrations.Migration):

    dependencies = [
        ('currency', '0001_initial'),
        ('io', '0008_importsource_resource_kwargs'),
        ('wbfdm', '0030_alter_relatedinstrumentthroughmodel_related_type'),
        ('wbportfolio', '0076_alter_dividendtransaction_price_and_more'),
    ]

    operations = [
        migrations.AlterField(
            model_name='transaction',
            name='transaction_type',
            field=models.CharField(
                choices=[('Trade', 'Trade'), ('DividendTransaction', 'Dividend Transaction'), ('Fees', 'Fees')],
                default='Trade', max_length=255, verbose_name='Type'),
        ),
        migrations.DeleteModel(
            name='Expiry',
        ),
        migrations.RemoveIndex(
            model_name='fees',
            name='wbportfolio_linked__a49c5d_idx',
        ),
        migrations.RemoveIndex(
            model_name='fees',
            name='wbportfolio_transac_1f7a29_idx',
        ),
        migrations.RenameField(
            model_name='fees',
            old_name='linked_product',
            new_name='product',
        ),
        migrations.RenameField(
            model_name='trade',
            old_name='external_identifier2',
            new_name='external_id_alternative',
        ),
        SetModelBasesOptionOperation("Fees", (models.Model, )),
        SetModelBasesOptionOperation("Trade", (models.Model, )),
        SetModelBasesOptionOperation("DividendTransaction", (models.Model, )),
        migrations.RenameField(
            model_name='dividendtransaction',
            old_name='transaction_ptr',
            new_name='id',
        ),
        migrations.RenameField(
            model_name='fees',
            old_name='transaction_ptr',
            new_name='id',
        ),
        migrations.RenameField(
            model_name='trade',
            old_name='transaction_ptr',
            new_name='id',
        ),
        migrations.AlterField(
            model_name='dividendtransaction',
            name='id',
            field=models.AutoField(auto_created=True, primary_key=True, serialize=False,
                                   verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='fees',
            name='id',
            field=models.AutoField(auto_created=True, primary_key=True, serialize=False,
                                   verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='trade',
            name='id',
            field=models.AutoField(auto_created=True, primary_key=True, serialize=False,
                                   verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='dividendtransaction',
            name='price',
            field=models.DecimalField(decimal_places=4, help_text='The amount paid per share', max_digits=15,
                                      verbose_name='DPS'),
        ),
        migrations.AlterField(
            model_name='dividendtransaction',
            name='shares',
            field=models.DecimalField(decimal_places=4, default=Decimal('0.0'),
                                      help_text='The number of shares held at record date, used to calculate the dividend',
                                      max_digits=15, verbose_name='Shares / Quantity'),
        ),
        migrations.AlterField(
            model_name='fees',
            name='fee_date',
            field=models.DateField(help_text='The date that this fee was paid.', verbose_name='Fees Date'),
        ),


        migrations.AddField(
            model_name='dividendtransaction',
            name='comment',
            field=models.TextField(blank=True, default='', verbose_name='Comment'),
        ),
        migrations.AddField(
            model_name='dividendtransaction',
            name='created',
            field=models.DateTimeField(auto_now_add=True, default=django.utils.timezone.now),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='dividendtransaction',
            name='currency',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, related_name='%(class)ss', to='currency.currency', verbose_name='Currency'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='dividendtransaction',
            name='currency_fx_rate',
            field=models.DecimalField(decimal_places=8, default=Decimal('1'), max_digits=14, verbose_name='FOREX rate'),
        ),
        migrations.AddField(
            model_name='dividendtransaction',
            name='distribution_method',
            field=models.CharField(choices=[('Payment', 'Payment'), ('Reinvestment', 'Reinvestment')], default='Payment', max_length=255, verbose_name='Type'),
        ),
        migrations.AddField(
            model_name='dividendtransaction',
            name='ex_date',
            field=models.DateField(blank=True, null=True, help_text='The date on which the stock starts trading without the dividend', verbose_name='Ex-Dividend Date'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='dividendtransaction',
            name='import_source',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='io.importsource'),
        ),
        migrations.AddField(
            model_name='dividendtransaction',
            name='portfolio',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, related_name='%(class)ss', to='wbportfolio.portfolio', verbose_name='Portfolio'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='dividendtransaction',
            name='record_date',
            field=models.DateField(blank=True, null=True, help_text='The date on which the holder must own the shares to be eligible for the dividend', verbose_name='Record Date'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='dividendtransaction',
            name='underlying_instrument',
            field=models.ForeignKey(blank=True, null=True, help_text='The instrument that is this transaction.', limit_choices_to=models.Q(('children__isnull', True)), on_delete=django.db.models.deletion.PROTECT, related_name='%(class)ss', to='wbfdm.instrument', verbose_name='Underlying Instrument'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='dividendtransaction',
            name='updated',
            field=models.DateTimeField(auto_now=True),
        ),
        migrations.AddField(
            model_name='dividendtransaction',
            name='value_date',
            field=models.DateField(blank=True, null=True, help_text='The date that this transaction was valuated/paid.', verbose_name='Value Date'),
            preserve_default=False,
        ),

        migrations.AddField(
            model_name='fees',
            name='currency',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT,
                                    related_name='fees', to='currency.currency', verbose_name='Currency'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='fees',
            name='import_source',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL,
                                    to='io.importsource'),
        ),
        migrations.AddField(
            model_name='fees',
            name='currency_fx_rate',
            field=models.DecimalField(decimal_places=8, default=Decimal('1'), max_digits=14, verbose_name='FOREX rate'),
        ),
        migrations.AddField(
            model_name='fees',
            name='created',
            field=models.DateTimeField(auto_now_add=True, default=django.utils.timezone.now),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='fees',
            name='total_value',
            field=models.DecimalField(decimal_places=4, blank=True, null=True, max_digits=20, verbose_name='Total Value'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='fees',
            name='total_value_gross',
            field=models.DecimalField(decimal_places=4, blank=True, null=True, max_digits=20, verbose_name='Total Value Gross'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='fees',
            name='updated',
            field=models.DateTimeField(auto_now=True),
        ),


        migrations.AddField(
            model_name='trade',
            name='currency_fx_rate',
            field=models.DecimalField(decimal_places=8, default=Decimal('1'), max_digits=14, verbose_name='FOREX rate'),
        ),
        migrations.AddField(
            model_name='trade',
            name='book_date',
            field=models.DateField(blank=True, null=True, help_text='The date that this transaction was booked.', verbose_name='Trade Date'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='trade',
            name='comment',
            field=models.TextField(blank=True, default='', verbose_name='Comment'),
        ),
        migrations.AddField(
            model_name='trade',
            name='created',
            field=models.DateTimeField(auto_now_add=True, default=django.utils.timezone.now),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='trade',
            name='external_id',
            field=models.CharField(blank=True, help_text='An external identifier that was supplied.', max_length=255, null=True, verbose_name='External Identifier'),
        ),
        migrations.AddField(
            model_name='trade',
            name='transaction_date',
            field=models.DateField(blank=True, null=True, help_text='The date that this transaction was traded.', verbose_name='Trade Date'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='trade',
            name='updated',
            field=models.DateTimeField(auto_now=True),
        ),
        migrations.AddField(
            model_name='trade',
            name='value_date',
            field=models.DateField(blank=True, null=True, help_text='The date that this transaction was valuated/paid.', verbose_name='Value Date'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='trade',
            name='currency',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT,
                                    related_name='%(class)ss', to='currency.currency', verbose_name='Currency'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='trade',
            name='import_source',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL,
                                    to='io.importsource'),
        ),
        migrations.AddField(
            model_name='trade',
            name='portfolio',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT,
                                    related_name='%(class)ss', to='wbportfolio.portfolio', verbose_name='Portfolio'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='trade',
            name='underlying_instrument',
            field=models.ForeignKey(blank=True, null=True, help_text='The instrument that is this transaction.',
                                    limit_choices_to=models.Q(('children__isnull', True)),
                                    on_delete=django.db.models.deletion.PROTECT, related_name='%(class)ss',
                                    to='wbfdm.instrument', verbose_name='Underlying Instrument'),
            preserve_default=False,
        ),


        migrations.RunSQL(
            """
            UPDATE wbportfolio_dividendtransaction AS o
            SET comment = t.comment, currency_id = t.currency_id, currency_fx_rate = t.currency_fx_rate, ex_date = t.transaction_date, import_source_id = t.import_source_id, portfolio_id = t.portfolio_id, record_date = t.transaction_date, underlying_instrument_id = t.underlying_instrument_id, value_date = t.value_date
            FROM wbportfolio_transaction AS t
            WHERE o.id = t.id
            """
        ),
        migrations.RunSQL(
            """
            UPDATE wbportfolio_fees AS o
            SET currency_id = t.currency_id, import_source_id = t.import_source_id, currency_fx_rate = t.currency_fx_rate, total_value = t.total_value, total_value_gross = t.total_value_gross
            FROM wbportfolio_transaction AS t
            WHERE o.id = t.id
            """
        ),
        migrations.RunSQL(
            """
            UPDATE wbportfolio_trade o
            SET underlying_instrument_id = t.underlying_instrument_id, portfolio_id = t.portfolio_id, currency_fx_rate = t.currency_fx_rate, book_date = t.book_date, comment = t.comment, external_id = t.external_id, transaction_date = t.transaction_date, value_date = t.value_date, currency_id = t.currency_id, import_source_id = t.import_source_id
            FROM wbportfolio_transaction AS t
            WHERE o.id = t.id
            """
        ),



        migrations.AlterField(
            model_name='dividendtransaction',
            name='currency',
            field=models.ForeignKey(default=None, on_delete=django.db.models.deletion.PROTECT, related_name='%(class)ss',
                                    to='currency.currency', verbose_name='Currency'),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='dividendtransaction',
            name='ex_date',
            field=models.DateField(default=None,
                                   help_text='The date on which the stock starts trading without the dividend',
                                   verbose_name='Ex-Dividend Date'),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='dividendtransaction',
            name='portfolio',
            field=models.ForeignKey(default=None, on_delete=django.db.models.deletion.PROTECT, related_name='%(class)ss',
                                    to='wbportfolio.portfolio', verbose_name='Portfolio'),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='dividendtransaction',
            name='record_date',
            field=models.DateField(default=None,
                                   help_text='The date on which the holder must own the shares to be eligible for the dividend',
                                   verbose_name='Record Date'),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='dividendtransaction',
            name='underlying_instrument',
            field=models.ForeignKey(default=None, help_text='The instrument that is this transaction.',
                                    limit_choices_to=models.Q(('children__isnull', True)),
                                    on_delete=django.db.models.deletion.PROTECT, related_name='%(class)ss',
                                    to='wbfdm.instrument', verbose_name='Underlying Instrument'),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='dividendtransaction',
            name='value_date',
            field=models.DateField(default=None, help_text='The date that this transaction was valuated/paid.',
                                   verbose_name='Value Date'),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='fees',
            name='currency',
            field=models.ForeignKey(default=None, on_delete=django.db.models.deletion.PROTECT, related_name='fees',
                                    to='currency.currency', verbose_name='Currency'),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='fees',
            name='total_value',
            field=models.DecimalField(decimal_places=4, default=None, max_digits=20, verbose_name='Total Value'),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='fees',
            name='total_value_gross',
            field=models.DecimalField(decimal_places=4, default=None, max_digits=20, verbose_name='Total Value Gross'),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='fees',
            name='product',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='fees',
                                    to='wbportfolio.product', verbose_name='Product'),
        ),
        migrations.AlterField(
            model_name='trade',
            name='book_date',
            field=models.DateField(default=None, help_text='The date that this transaction was booked.',
                                   verbose_name='Trade Date'),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='trade',
            name='currency',
            field=models.ForeignKey(default=None, on_delete=django.db.models.deletion.PROTECT, related_name='%(class)ss',
                                    to='currency.currency', verbose_name='Currency'),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='trade',
            name='external_id_alternative',
            field=models.CharField(blank=True, help_text='A second external identifier that was supplied.',
                                   max_length=255, null=True, verbose_name='Alternative External Identifier'),
        ),
        migrations.AlterField(
            model_name='trade',
            name='portfolio',
            field=models.ForeignKey(default=None, on_delete=django.db.models.deletion.PROTECT, related_name='%(class)ss',
                                    to='wbportfolio.portfolio', verbose_name='Portfolio'),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='trade',
            name='transaction_date',
            field=models.DateField(default=None, help_text='The date that this transaction was traded.',
                                   verbose_name='Trade Date'),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='trade',
            name='underlying_instrument',
            field=models.ForeignKey(default=None, help_text='The instrument that is this transaction.',
                                    limit_choices_to=models.Q(('children__isnull', True)),
                                    on_delete=django.db.models.deletion.PROTECT, related_name='%(class)ss',
                                    to='wbfdm.instrument', verbose_name='Underlying Instrument'),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='trade',
            name='value_date',
            field=models.DateField(default=None, help_text='The date that this transaction was valuated/paid.',
                                   verbose_name='Value Date'),
            preserve_default=False,
        ),

        migrations.AddField(
            model_name='dividendtransaction',
            name='fees',
            field=models.GeneratedField(db_persist=True, expression=django.db.models.expressions.CombinedExpression(
                models.F('price_gross'), '-', models.F('price')),
                                        output_field=models.DecimalField(decimal_places=4, max_digits=20)),
        ),
        migrations.AddField(
            model_name='dividendtransaction',
            name='total_value',
            field=models.GeneratedField(db_persist=True,
                                        expression=django.db.models.expressions.CombinedExpression(models.F('price'),
                                                                                                   '*',
                                                                                                   models.F('shares')),
                                        output_field=models.DecimalField(decimal_places=4, max_digits=20)),
        ),
        migrations.AddField(
            model_name='dividendtransaction',
            name='total_value_fx_portfolio',
            field=models.GeneratedField(db_persist=True, expression=django.db.models.expressions.CombinedExpression(
                django.db.models.expressions.CombinedExpression(models.F('currency_fx_rate'), '*', models.F('price')),
                '*', models.F('shares')), output_field=models.DecimalField(decimal_places=4, max_digits=20)),
        ),
        migrations.AddField(
            model_name='dividendtransaction',
            name='total_value_gross',
            field=models.GeneratedField(db_persist=True, expression=django.db.models.expressions.CombinedExpression(
                django.db.models.expressions.CombinedExpression(models.F('price'), '*', models.F('shares')), '*',
                models.F('retrocession')), output_field=models.DecimalField(decimal_places=4, max_digits=20)),
        ),
        migrations.AddField(
            model_name='dividendtransaction',
            name='total_value_gross_fx_portfolio',
            field=models.GeneratedField(db_persist=True, expression=django.db.models.expressions.CombinedExpression(
                django.db.models.expressions.CombinedExpression(models.F('currency_fx_rate'), '*',
                                                                models.F('price_gross')), '*', models.F('shares')),
                                        output_field=models.DecimalField(decimal_places=4, max_digits=20)),
        ),
        migrations.AddField(
            model_name='trade',
            name='total_value',
            field=models.GeneratedField(db_persist=True,
                                        expression=django.db.models.expressions.CombinedExpression(models.F('price'),
                                                                                                   '*',
                                                                                                   models.F('shares')),
                                        output_field=models.DecimalField(decimal_places=4, max_digits=20)),
        ),
        migrations.AddField(
            model_name='trade',
            name='total_value_gross',
            field=models.GeneratedField(db_persist=True, expression=django.db.models.expressions.CombinedExpression(
                models.F('price_gross'), '*', models.F('shares')),
                                        output_field=models.DecimalField(decimal_places=4, max_digits=20)),
        ),
        migrations.AddField(
            model_name='trade',
            name='fees',
            field=models.GeneratedField(db_persist=True, expression=django.db.models.expressions.CombinedExpression(
                models.F('price_gross'), '-', models.F('price')),
                                        output_field=models.DecimalField(decimal_places=4, max_digits=20)),
        ),
        migrations.AddField(
            model_name='trade',
            name='total_value_fx_portfolio',
            field=models.GeneratedField(db_persist=True, expression=django.db.models.expressions.CombinedExpression(
                django.db.models.expressions.CombinedExpression(models.F('currency_fx_rate'), '*', models.F('price')),
                '*', models.F('shares')), output_field=models.DecimalField(decimal_places=4, max_digits=20)),
        ),
        migrations.AddField(
            model_name='trade',
            name='total_value_gross_fx_portfolio',
            field=models.GeneratedField(db_persist=True, expression=django.db.models.expressions.CombinedExpression(
                django.db.models.expressions.CombinedExpression(models.F('currency_fx_rate'), '*',
                                                                models.F('price_gross')), '*', models.F('shares')),
                                        output_field=models.DecimalField(decimal_places=4, max_digits=20)),
        ),
        migrations.AlterField(
            model_name='trade',
            name='price',
            field=models.DecimalField(decimal_places=4, help_text='The price per share.', max_digits=16,
                                      verbose_name='Price'),
        ),
        migrations.AddField(
            model_name='fees',
            name='total_value_fx_portfolio',
            field=models.GeneratedField(db_persist=True, expression=django.db.models.expressions.CombinedExpression(
                models.F('currency_fx_rate'), '*', models.F('total_value')),
                                        output_field=models.DecimalField(decimal_places=4, max_digits=20)),
        ),
        migrations.AddField(
            model_name='fees',
            name='total_value_gross_fx_portfolio',
            field=models.GeneratedField(db_persist=True, expression=django.db.models.expressions.CombinedExpression(
                models.F('currency_fx_rate'), '*', models.F('total_value_gross')),
                                        output_field=models.DecimalField(decimal_places=4, max_digits=20)),
        ),

        # cleanup
        migrations.RemoveField(
            model_name='transaction',
            name='currency',
        ),
        migrations.RemoveField(
            model_name='transaction',
            name='import_source',
        ),
        migrations.RemoveField(
            model_name='transaction',
            name='portfolio',
        ),
        migrations.RemoveField(
            model_name='transaction',
            name='underlying_instrument',
        ),
        migrations.DeleteModel(
            name='Transaction',
        ),
        migrations.AddIndex(
            model_name='fees',
            index=models.Index(fields=['product'], name='wbportfolio_product_daf107_idx'),
        ),
        migrations.AddIndex(
            model_name='fees',
            index=models.Index(fields=['transaction_subtype', 'product', 'fee_date', 'calculated'],
                               name='wbportfolio_transac_7355c2_idx'),
        ),
        migrations.AddIndex(
            model_name='trade',
            index=models.Index(fields=['underlying_instrument', 'transaction_date'],
                               name='wbportfolio_underly_a0f0ff_idx'),
        ),
        migrations.AddIndex(
            model_name='trade',
            index=models.Index(fields=['portfolio', 'underlying_instrument', 'transaction_date'],
                               name='wbportfolio_portfol_6a42a2_idx'),
        ),
        migrations.RunPython(remove_duplicate),
        migrations.AddConstraint(
            model_name='trade',
            constraint=models.UniqueConstraint(condition=models.Q(('trade_proposal__isnull', False)),
                                               fields=('portfolio', 'transaction_date', 'underlying_instrument'),
                                               name='unique_manual_trade'),
        ),
    ]



from datetime import date, datetime

import numpy as np
import pandas as pd
import plotly.graph_objects as go
from wbcore import viewsets
from wbcrm.models.accounts import Account

from wbportfolio.analysis.claims import get_assets_and_net_new_money_progression
from wbportfolio.filters import AssetsAndNetNewMoneyProgressionFilterSet
from wbportfolio.models.products import Product
from wbportfolio.models.transactions.claim import Claim

from .configs import AssetAndNetNewMoneyProgressionChartTitleConfig


class AssetAndNetNewMoneyProgressionChartViewSet(viewsets.ChartViewSet):
    queryset = Claim.objects.all()

    filterset_class = AssetsAndNetNewMoneyProgressionFilterSet

    title_config_class = AssetAndNetNewMoneyProgressionChartTitleConfig

    def get_plotly(self, queryset) -> "go.Figure":
        if "period" not in self.request.GET:
            from_date = date(2023, 12, 31)
            to_date = date(2024, 6, 30)
        else:
            from_date, to_date = self.request.GET["period"].split(",")
            from_date = datetime.strptime(from_date, "%Y-%m-%d").date()
            to_date = datetime.strptime(to_date, "%Y-%m-%d").date()

        account_id = self.request.GET.get("account", None)
        product_id = self.request.GET.get("product", None)

        name = f"Progression: {from_date:%d.%m.%Y}-{to_date:%d.%m.%Y}"

        account_tree_id = None
        el = list()
        if account_id:
            account = Account.objects.get(id=account_id)
            account_tree_id = account.tree_id  # type: ignore
            el.append(str(account))

        if product_id:
            el.append(str(Product.objects.get(id=product_id)))

        if len(el) > 0:
            name += f" ({", ".join(el)})"

        data = get_assets_and_net_new_money_progression(
            from_date=from_date, to_date=to_date, account_tree_id=account_tree_id, product_id=product_id
        )
        df = pd.DataFrame(data)

        # Compute performance of each month
        df["performance"] = df["total_assets"] - df["total_assets"].shift(1) - df["net_new_money"]

        # Compute the x label
        df["x"] = df[["year", "month"]].apply(lambda row: "-".join(row.values.astype(str)), axis=1)

        # Implode the pivoted table to have a list of values
        df = df.melt(id_vars=["x", "year", "month"], value_vars=["net_new_money", "performance", "total_assets"])

        # Sort by year, then month, then variable to have everything in the right order
        df = df.sort_values(["year", "month", "variable"])

        # Remove the first two rows, as they are the NNM and Performance of the previous month, which we do not want to display
        df = df.iloc[2:]

        # Create a pattern to have the measure for the waterfall chart
        df["measure"] = np.tile(["absolute", "relative", "relative"], len(df) // 3 + 1)[: len(df)]

        # We reset the index and drop all columns we don't need
        # IMPORTANT: Do not drop year and month, even though we
        # don't need it - we will loose the ordering
        df = df.reset_index()[["x", "year", "month", "variable", "value", "measure"]]

        # Rename things, to have it pretty in the chart
        df["variable"] = (
            df["variable"]
            .str.replace("net_new_money", "NNM")
            .replace("total_assets", "AuM")
            .replace("performance", "Performance")
        )
        df.iloc[0, 0] = ""
        df.iloc[0, 3] = "Initial AuM"
        df.iloc[-1, 3] = "Final AuM"

        # Create the figure and show it
        fig = go.Figure()
        fig.add_trace(
            go.Waterfall(
                name=name,
                orientation="v",
                measure=df["measure"],
                x=[df["x"], df["variable"]],
                y=df["value"],
                text=list(map(lambda x: f"{x/1_000_000:,.0f}" if x is not None else "0", df["value"])),
                decreasing={"marker": {"color": "#FF6961"}},
                increasing={"marker": {"color": "#77DD77"}},
                totals={"marker": {"color": "#D3D3D3"}},
            )
        )
        fig.update_layout(
            template="plotly_white",
            margin=dict(l=10, r=10, t=0, b=40),
            showlegend=True,
            legend={
                "orientation": "h",
                "yanchor": "bottom",
                "y": 1,
                "xanchor": "center",
                "x": 0.5,
            },
        )

        return fig



from typing import Optional

from wbcore.contrib.color.enums import WBColor
from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display.instance_display import Inline, Layout, Page
from wbcore.metadata.configs.display.instance_display.operators import default
from wbcore.metadata.configs.display.instance_display.shortcuts import Display
from wbcore.metadata.configs.display.view_config import DisplayViewConfig

from wbportfolio.models import OrderProposal


class OrderProposalDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="trade_date", label="Order Date"),
                dp.Field(key="comment", label="Comment"),
            ],
            legends=[
                dp.Legend(
                    key="status",
                    items=[
                        dp.LegendItem(
                            icon=WBColor.BLUE_LIGHT.value,
                            label=OrderProposal.Status.DRAFT.label,
                            value=OrderProposal.Status.DRAFT.value,
                        ),
                        dp.LegendItem(
                            icon=WBColor.YELLOW_LIGHT.value,
                            label=OrderProposal.Status.SUBMIT.label,
                            value=OrderProposal.Status.SUBMIT.value,
                        ),
                        dp.LegendItem(
                            icon=WBColor.GREEN_LIGHT.value,
                            label=OrderProposal.Status.APPROVED.label,
                            value=OrderProposal.Status.APPROVED.value,
                        ),
                        dp.LegendItem(
                            icon=WBColor.RED_LIGHT.value,
                            label=OrderProposal.Status.DENIED.label,
                            value=OrderProposal.Status.DENIED.value,
                        ),
                        dp.LegendItem(
                            icon=WBColor.RED_DARK.value,
                            label=OrderProposal.Status.FAILED.label,
                            value=OrderProposal.Status.FAILED.value,
                        ),
                    ],
                ),
            ],
            formatting=[
                dp.Formatting(
                    column="status",
                    formatting_rules=[
                        dp.FormattingRule(
                            style={"backgroundColor": WBColor.BLUE_LIGHT.value},
                            condition=("==", OrderProposal.Status.DRAFT.value),
                        ),
                        dp.FormattingRule(
                            style={"backgroundColor": WBColor.YELLOW_LIGHT.value},
                            condition=("==", OrderProposal.Status.SUBMIT.value),
                        ),
                        dp.FormattingRule(
                            style={"backgroundColor": WBColor.GREEN_LIGHT.value},
                            condition=("==", OrderProposal.Status.APPROVED.value),
                        ),
                        dp.FormattingRule(
                            style={"backgroundColor": WBColor.RED_LIGHT.value},
                            condition=("==", OrderProposal.Status.DENIED.value),
                        ),
                        dp.FormattingRule(
                            style={"backgroundColor": WBColor.RED_DARK.value},
                            condition=("==", OrderProposal.Status.FAILED.value),
                        ),
                    ],
                )
            ],
        )

    def get_instance_display(self) -> Display:
        return Display(
            pages=[
                Page(
                    title="Main Information",
                    layouts={
                        default(): Layout(
                            grid_template_areas=[
                                ["status", "status"],
                                ["trade_date", "total_cash_weight"],
                                ["rebalancing_model", "target_portfolio"]
                                if self.view.new_mode
                                else ["rebalancing_model", "rebalancing_model"],
                                ["comment", "comment"],
                            ],
                        ),
                    },
                ),
                Page(
                    title="Orders",
                    layouts={
                        default(): Layout(
                            grid_template_areas=[["orders"]],
                            grid_template_rows=["1fr"],
                            inlines=[Inline(key="orders", endpoint="orders")],
                        ),
                    },
                ),
            ]
        )



from wbcore import serializers as wb_serializers

from wbportfolio.models import AssetPosition


class AggregatedAssetPositionModelSerializer(wb_serializers.ModelSerializer):
    id = wb_serializers.PrimaryKeyCharField()
    aggregated_title = wb_serializers.CharField(read_only=True)

    sum_total_value_start = wb_serializers.DecimalField(read_only=True, max_digits=16, decimal_places=4)
    sum_total_value = wb_serializers.DecimalField(read_only=True, max_digits=16, decimal_places=4)
    weighting = wb_serializers.DecimalField(
        read_only=True,
        max_digits=6,
        decimal_places=4,
        percent=True,
        decorators=[{"position": "right", "value": "%"}],
    )
    performance = wb_serializers.DecimalField(
        read_only=True,
        max_digits=6,
        decimal_places=4,
        percent=True,
        decorators=[{"position": "right", "value": "%"}],
    )
    performance_local = wb_serializers.DecimalField(
        read_only=True,
        max_digits=6,
        decimal_places=4,
        percent=True,
        decorators=[{"position": "right", "value": "%"}],
    )
    contribution = wb_serializers.DecimalField(
        read_only=True,
        max_digits=6,
        decimal_places=4,
        percent=True,
        decorators=[{"position": "right", "value": "%"}],
    )
    contribution_local = wb_serializers.DecimalField(
        read_only=True,
        max_digits=6,
        decimal_places=4,
        percent=True,
        decorators=[{"position": "right", "value": "%"}],
    )
    difference_performance = wb_serializers.DecimalField(
        read_only=True,
        max_digits=6,
        decimal_places=4,
        percent=True,
        decorators=[{"position": "right", "value": "%"}],
    )
    difference_contribution = wb_serializers.DecimalField(
        read_only=True,
        max_digits=6,
        decimal_places=4,
        percent=True,
        decorators=[{"position": "right", "value": "%"}],
    )

    class Meta:
        model = AssetPosition
        fields = (
            "id",
            "aggregated_title",
            "sum_total_value_start",
            "sum_total_value",
            "weighting",
            "performance",
            "performance_local",
            "contribution",
            "contribution_local",
            "difference_performance",
            "difference_contribution",
            "_additional_resources",
        )



from celery import shared_task
from tqdm import tqdm
from wbfdm.models import Controversy, Instrument

from wbportfolio.models import AssetPosition


@shared_task(queue="portfolio")
def synchronize_portfolio_controversies():
    qs = AssetPosition.objects.values("underlying_instrument").distinct("underlying_instrument")
    for row in tqdm(qs, total=qs.count()):
        for controversy in Instrument.objects.filter(id=row["underlying_instrument"]).dl.esg_controversies():
            Controversy.sync_from_dataloader(controversy)



from datetime import date

from wbportfolio.pms.typing import Portfolio as PortfolioDTO


class AbstractRebalancingModel:
    @property
    def validation_errors(self) -> str:
        return getattr(self, "_validation_errors", "Rebalacing cannot applied for these parameters")

    def __init__(
        self,
        portfolio,
        trade_date: date,
        last_effective_date: date,
        effective_portfolio: PortfolioDTO | None = None,
        **kwargs,
    ):
        self.portfolio = portfolio
        self.trade_date = trade_date
        self.last_effective_date = last_effective_date
        self.effective_portfolio = effective_portfolio
        # we try to get the portfolio at the trade date
        if not self.effective_portfolio:
            self.effective_portfolio = self.portfolio._build_dto(self.last_effective_date)

    def is_valid(self) -> bool:
        return True

    def get_target_portfolio(self) -> PortfolioDTO:
        raise NotImplementedError()



import logging
from datetime import date

from dateutil import rrule
from django.core.exceptions import ObjectDoesNotExist, ValidationError
from django.db import models
from django.db.models.signals import post_migrate
from django.dispatch import receiver
from django.utils.functional import cached_property
from django.utils.module_loading import autodiscover_modules
from django.utils.translation import gettext_lazy as _
from pandas._libs.tslibs.offsets import BDay
from wbcore.utils.importlib import import_from_dotted_path
from wbcore.utils.models import ComplexToStringMixin
from wbcore.utils.rrules import convert_rrulestr_to_dict, humanize_rrule

from wbportfolio.models.orders.order_proposals import OrderProposal
from wbportfolio.models.portfolio import Portfolio
from wbportfolio.pms.typing import Portfolio as PortfolioDTO
from wbportfolio.rebalancing.base import AbstractRebalancingModel

logger = logging.getLogger("pms")


class RebalancingModel(models.Model):
    name = models.CharField(max_length=64, verbose_name="Name")
    class_path = models.CharField(max_length=512, verbose_name="Class path")

    def __str__(self):
        return self.name

    class Meta:
        verbose_name = "Rebalancing Model"
        verbose_name_plural = "Rebalancing Models"

    @cached_property
    def model_class(self) -> type[AbstractRebalancingModel]:
        """
        Return the imported backend class
        Returns:
            The backend class
        """
        return import_from_dotted_path(self.class_path)

    def get_target_portfolio(
        self,
        portfolio: Portfolio,
        trade_date: date,
        last_effective_date: date,
        effective_portfolio: PortfolioDTO | None = None,
        **kwargs,
    ) -> PortfolioDTO:
        model = self.model_class(
            portfolio, trade_date, last_effective_date, effective_portfolio=effective_portfolio, **kwargs
        )
        if not model.is_valid():
            raise ValidationError(model.validation_errors)
        return model.get_target_portfolio()

    @classmethod
    def get_representation_endpoint(cls) -> str:
        return "wbportfolio:rebalancingmodelrepresentation-list"

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"

    @classmethod
    def get_representation_label_key(cls) -> str:
        return "{{name}}"


class Rebalancer(ComplexToStringMixin, models.Model):
    portfolio = models.OneToOneField(
        "wbportfolio.Portfolio", on_delete=models.CASCADE, related_name="automatic_rebalancer"
    )
    rebalancing_model = models.ForeignKey(
        RebalancingModel, on_delete=models.PROTECT, related_name="rebalancers", verbose_name="Rebalancing Model"
    )
    parameters = models.JSONField(default=dict, verbose_name="Parameters", blank=True)
    approve_order_proposal_automatically = models.BooleanField(
        default=False, verbose_name="Apply Order Proposal Automatically"
    )
    activation_date = models.DateField(verbose_name="Activation Date")
    frequency = models.CharField(
        default="RRULE:FREQ=MONTHLY;BYDAY=MO,TU,WE,TH,FR;BYSETPOS=1",
        max_length=256,
        verbose_name=_("Evaluation Frequency"),
        help_text=_("The Evaluation Frequency in RRULE format"),
    )

    def save(self, *args, **kwargs):
        if not self.activation_date:
            try:
                self.activation_date = self.portfolio.assets.earliest("date").date
            except ObjectDoesNotExist:
                self.activation_date = date.today()
        super().save(*args, **kwargs)

    def _get_next_valid_date(self, valid_date: date) -> date:
        pivot_date = valid_date
        while OrderProposal.objects.filter(
            portfolio=self.portfolio, status=OrderProposal.Status.FAILED, trade_date=pivot_date
        ).exists():
            pivot_date = (pivot_date + BDay(1)).date()
        return pivot_date

    def is_valid(self, trade_date: date) -> bool:
        if OrderProposal.objects.filter(
            portfolio=self.portfolio,
            status=OrderProposal.Status.APPROVED,
            trade_date=trade_date,
            rebalancing_model__isnull=True,
        ).exists():  # if a already approved order proposal exists, we do not allow a re-evaluatioon of the rebalancing (only possible if "replayed")
            return False
        for initial_valid_datetime in self.get_rrule(trade_date):
            initial_valid_date = initial_valid_datetime.date()
            alternative_valid_date = self._get_next_valid_date(initial_valid_date)
            if trade_date in [alternative_valid_date, initial_valid_date]:
                return True
            if alternative_valid_date > trade_date:
                break
        return False

    def evaluate_rebalancing(self, trade_date: date, effective_portfolio=None):
        order_proposal, _ = OrderProposal.objects.get_or_create(
            trade_date=trade_date,
            portfolio=self.portfolio,
            defaults={
                "comment": "Automatic rebalancing",
                "rebalancing_model": self.rebalancing_model,
            },
        )

        if order_proposal.rebalancing_model == self.rebalancing_model:
            try:
                logger.info(
                    f"Getting target portfolio ({self.portfolio}) for rebalancing model {self.rebalancing_model} for trade date {trade_date:%Y-%m-%d}"
                )
                target_portfolio = self.rebalancing_model.get_target_portfolio(
                    self.portfolio,
                    order_proposal.trade_date,
                    order_proposal.value_date,
                    effective_portfolio=effective_portfolio,
                    **self.parameters,
                )
                order_proposal.approve_workflow(
                    approve_automatically=self.approve_order_proposal_automatically,
                    target_portfolio=target_portfolio,
                    effective_portfolio=effective_portfolio,
                )
            except ValidationError as e:
                logger.warning(f"Validation error while approving the orders: {e}")
                # If we encountered a validation error, we set the order proposal as failed
                order_proposal.status = OrderProposal.Status.FAILED
            order_proposal.save()
        return order_proposal

    @property
    def rrule(self):
        return self.get_rrule()

    def get_next_rebalancing_date(self, pivot_date: date) -> date | None:
        for _dt in self.rrule:
            _d = _dt.date()
            if _d > pivot_date:
                return _d

    @property
    def frequency_repr(self):
        return humanize_rrule(self.rrule)

    def get_rrule(self, to_date: date | None = None, count: int | None = None):
        rrule_dict = convert_rrulestr_to_dict(self.frequency, dtstart=self.activation_date, until=to_date, count=count)
        return rrule.rrule(**rrule_dict)

    def compute_str(self):
        return f"{self.frequency_repr} {self.portfolio.name} ({self.rebalancing_model.name})"

    class Meta:
        verbose_name = "Rebalancer"
        verbose_name_plural = "Rebalancers"

    @classmethod
    def get_endpoint_basename(cls) -> str:
        return "wbportfolio:rebalancer"

    @classmethod
    def get_representation_endpoint(cls) -> str:
        return "wbportfolio:rebalancerrepresentation-list"

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"


@receiver(post_migrate, sender=RebalancingModel)
def post_migrate_rebalancing_model(sender, verbosity, interactive, stdout, using, plan, apps, **kwargs):
    autodiscover_modules("rebalancing.models")



from wbcore import serializers as wb_serializers
from wbcore.contrib.icons import WBIcon
from wbcore.enums import RequestType
from wbcore.metadata.configs import buttons as bt
from wbcore.metadata.configs.buttons.view_config import ButtonViewConfig
from wbcore.metadata.configs.display import create_simple_display


class NormalizeSerializer(wb_serializers.Serializer):
    total_cash_weight = wb_serializers.FloatField(default=0, precision=4, percent=True)


class OrderProposalButtonConfig(ButtonViewConfig):
    def get_custom_list_instance_buttons(self):
        return {
            bt.DropDownButton(
                label="Tools",
                buttons=(
                    bt.ActionButton(
                        method=RequestType.PATCH,
                        identifiers=("wbportfolio:orderproposal",),
                        key="replay",
                        icon=WBIcon.SYNCHRONIZE.icon,
                        label="Replay Orders",
                        description_fields="""
                        <p>Replay Orders. It will recompute all assets positions until next order proposal day (or today otherwise) </p>
                        """,
                        action_label="Replay Order",
                        title="Replay Order",
                    ),
                    bt.ActionButton(
                        method=RequestType.PATCH,
                        identifiers=("wbportfolio:orderproposal",),
                        key="reset",
                        icon=WBIcon.REGENERATE.icon,
                        label="Reset Orders",
                        description_fields="""
                            <p><strong>Warning:</strong> This action will delete all current orders and recreate initial orders based on your last effective portfolio.</p>
                            <p><strong>Note:</strong> All delta weights will be permanently removed. This operation cannot be undone.</p>
                            """,
                        action_label="Reset Orders",
                        title="Reset Orders",
                    ),
                    bt.ActionButton(
                        method=RequestType.PATCH,
                        identifiers=("wbportfolio:orderproposal",),
                        key="normalize",
                        icon=WBIcon.EDIT.icon,
                        label="Normalize Orders",
                        description_fields="""
                            <p>Make sure all orders normalize to a total target weight of (100 - {{total_cash_weight}})%</p>
                            """,
                        action_label="Normalize Orders",
                        title="Normalize Orders",
                        serializer=NormalizeSerializer,
                        instance_display=create_simple_display([["total_cash_weight"]]),
                    ),
                    bt.ActionButton(
                        method=RequestType.PATCH,
                        identifiers=("wbportfolio:orderproposal",),
                        key="deleteall",
                        icon=WBIcon.DELETE.icon,
                        label="Delete All Orders",
                        description_fields="""
                    <p>Delete all orders from this order proposal?</p>
                    """,
                        action_label="Delete All Orders",
                        title="Delete All Orders",
                    ),
                ),
            ),
        }

    def get_custom_instance_buttons(self):
        return self.get_custom_list_instance_buttons()



def get_perf_fee_isin(source):
    default_perf_fee_isin = "CH0040602242"
    if source:
        return source.import_parameters.get("performance_fees_instrument_isin", default_perf_fee_isin)
    return default_perf_fee_isin



from contextlib import suppress
from datetime import date
from decimal import Decimal

from django.contrib.messages import info, warning
from django.shortcuts import get_object_or_404
from django.utils.functional import cached_property
from pandas._libs.tslibs.offsets import BDay
from rest_framework import status
from rest_framework.decorators import action
from rest_framework.response import Response
from wbcompliance.viewsets.risk_management.mixins import RiskCheckViewSetMixin
from wbcore import serializers as wb_serializers
from wbcore import viewsets
from wbcore.metadata.configs.display.instance_display import (
    Display,
    create_simple_display,
)
from wbcore.permissions.permissions import InternalUserPermissionMixin
from wbcore.utils.views import CloneMixin

from wbportfolio.models import AssetPosition, OrderProposal
from wbportfolio.models.orders.order_proposals import (
    replay_as_task,
)
from wbportfolio.serializers import (
    OrderProposalModelSerializer,
    OrderProposalRepresentationSerializer,
    ReadOnlyOrderProposalModelSerializer,
)

from ..mixins import UserPortfolioRequestPermissionMixin
from .configs import (
    OrderProposalButtonConfig,
    OrderProposalDisplayConfig,
    OrderProposalEndpointConfig,
    OrderProposalPortfolioEndpointConfig,
)


class OrderProposalRepresentationViewSet(InternalUserPermissionMixin, viewsets.RepresentationViewSet):
    IDENTIFIER = "wbportfolio:trade"
    queryset = OrderProposal.objects.all()
    serializer_class = OrderProposalRepresentationSerializer


class OrderProposalModelViewSet(CloneMixin, RiskCheckViewSetMixin, InternalUserPermissionMixin, viewsets.ModelViewSet):
    ordering_fields = ("trade_date",)
    ordering = ("-trade_date",)
    search_fields = ("comment",)
    filterset_fields = {"trade_date": ["exact", "gte", "lte"], "status": ["exact"]}

    queryset = OrderProposal.objects.select_related("rebalancing_model", "portfolio")
    serializer_class = OrderProposalModelSerializer
    display_config_class = OrderProposalDisplayConfig
    button_config_class = OrderProposalButtonConfig
    endpoint_config_class = OrderProposalEndpointConfig

    def get_serializer_class(self):
        if self.new_mode or (
            "pk" in self.kwargs and (obj := self.get_object()) and obj.status == OrderProposal.Status.DRAFT
        ):
            return OrderProposalModelSerializer
        return ReadOnlyOrderProposalModelSerializer

    # 2 methods to parametrize the clone button functionality
    def get_clone_button_serializer_class(self, instance):
        class CloneSerializer(wb_serializers.Serializer):
            clone_date = wb_serializers.DateField(
                default=(instance.trade_date + BDay(1)).date(), label="Trade Date"
            )  # we need to change the field name from the trade proposa fields, otherwise fontend conflicts
            clone_comment = wb_serializers.TextField(label="Comment")

        return CloneSerializer

    def get_clone_button_instance_display(self) -> Display:
        return create_simple_display(
            [
                ["clone_comment"],
                ["clone_date"],
            ]
        )

    def add_messages(self, request, instance=None, **kwargs):
        if instance and instance.status == OrderProposal.Status.SUBMIT:
            if not instance.portfolio.is_manageable:
                info(request, "This order proposal cannot be approved the portfolio is considered unmanaged.")
            if instance.has_non_successful_checks:
                warning(
                    request,
                    "This order proposal cannot be approved because there is unsuccessful pre-trade checks. Please rectify accordingly and resubmit a valid order proposal",
                )

    @classmethod
    def _get_risk_checks_button_title(cls) -> str:
        return "Pre-Trade Checks"

    @action(detail=True, methods=["PATCH"])
    def reset(self, request, pk=None):
        order_proposal = get_object_or_404(OrderProposal, pk=pk)
        if order_proposal.status == OrderProposal.Status.DRAFT:
            order_proposal.orders.all().update(weighting=0)
            order_proposal.reset_orders()
            return Response({"send": True})
        return Response({"status": "Order Proposal is not Draft"}, status=status.HTTP_400_BAD_REQUEST)

    @action(detail=True, methods=["PATCH"])
    def normalize(self, request, pk=None):
        order_proposal = get_object_or_404(OrderProposal, pk=pk)
        total_cash_weight = Decimal(request.data.get("total_cash_weight", Decimal("0.0")))
        if order_proposal.status == OrderProposal.Status.DRAFT:
            order_proposal.normalize_orders(total_target_weight=Decimal("1.0") - total_cash_weight)
            return Response({"send": True})
        return Response({"status": "Order Proposal is not Draft"}, status=status.HTTP_400_BAD_REQUEST)

    @action(detail=True, methods=["PATCH"])
    def replay(self, request, pk=None):
        order_proposal = get_object_or_404(OrderProposal, pk=pk)
        if order_proposal.portfolio.is_manageable:
            replay_as_task.delay(order_proposal.id, user_id=self.request.user.id)
            return Response({"send": True})
        return Response({"status": "Order Proposal is not Draft"}, status=status.HTTP_400_BAD_REQUEST)

    @action(detail=True, methods=["PATCH"])
    def deleteall(self, request, pk=None):
        order_proposal = get_object_or_404(OrderProposal, pk=pk)
        if order_proposal.status == OrderProposal.Status.DRAFT:
            order_proposal.orders.all().delete()
            return Response({"send": True})
        return Response({"status": "Order Proposal is not Draft"}, status=status.HTTP_400_BAD_REQUEST)


class OrderProposalPortfolioModelViewSet(UserPortfolioRequestPermissionMixin, OrderProposalModelViewSet):
    endpoint_config_class = OrderProposalPortfolioEndpointConfig

    @cached_property
    def default_trade_date(self) -> date | None:
        with suppress(AssetPosition.DoesNotExist):
            return (self.portfolio.assets.latest("date").date + BDay(1)).date()

    def get_queryset(self):
        return OrderProposal.objects.filter(portfolio=self.kwargs["portfolio_id"])



# Generated by Django 5.0.8 on 2024-10-01 14:43

from collections import defaultdict
from contextlib import suppress

from django.db import migrations, models
from tqdm import tqdm


def remove_duplicates(apps, schema_editor):
    changed_portfolios = defaultdict(set)
    from wbportfolio.models import Portfolio

    AssetPosition = apps.get_model("wbportfolio", "AssetPosition")
    for row in (
        AssetPosition.objects.values("portfolio", "date", "underlying_instrument", "portfolio_created")
        .annotate(c=models.Count("*"))
        .filter(c__gt=1)
    ):
        with suppress(Exception):
            qs = AssetPosition.objects.filter(
                date=row["date"],
                portfolio=row["portfolio"],
                underlying_instrument=row["underlying_instrument"],
                portfolio_created=row["portfolio_created"],
            )
            qs.filter(weighting=0).delete()
            base_p = qs.first()
            for a in qs.exclude(id=base_p.id):
                if (
                    abs(base_p.weighting / a.weighting - 1) < 0.02
                    or (a.import_source is None and a.portfolio_created is None)
                    or (a.portfolio.id == 117 and a.underlying_instrument.id == 214)
                ):
                    changed_portfolios[a.portfolio].add(a.date)
                    a.delete()
            if base_p.import_source and (
                qs.filter(import_source=base_p.import_source).count() > 1
                or (base_p.portfolio.id == 123 and base_p.underlying_instrument.id == 214)
            ):
                base_p.weighting += qs.aggregate(c=models.Sum("weighting"))["c"]
                base_p.shares += qs.aggregate(c=models.Sum("shares"))["c"]
                base_p.save()
                changed_portfolios[base_p.portfolio].add(base_p.date)
                qs.filter(import_source=base_p.import_source).delete()

    # Normalizing portfolio weights
    for portfolio, dates in tqdm(changed_portfolios.items()):
        portfolio = Portfolio.all_objects.get(id=portfolio.id)
        for _d in dates:
            portfolio.change_at_date(_d)


class Migration(migrations.Migration):
    dependencies = [
        ("currency", "0001_initial"),
        ("io", "0007_alter_exportsource_query_params"),
        ("wbfdm", "0026_instrument_is_cash_equivalent"),
        ("wbportfolio", "0066_assetposition_initial_shares_at_custodian_and_more"),
    ]

    operations = [
        migrations.RunSQL(
            sql="SET CONSTRAINTS ALL IMMEDIATE;",
        ),
        migrations.RunPython(remove_duplicates),
        migrations.RunSQL(
            sql="SET CONSTRAINTS ALL DEFERRED;",
        ),
        migrations.AddConstraint(
            model_name="assetposition",
            constraint=models.UniqueConstraint(
                fields=("portfolio", "date", "underlying_instrument", "portfolio_created"),
                name="unique_asset_position",
            ),
        ),
    ]



from .claim import (
    ClaimFilter,
    ClaimGroupByFilter,
    ConsolidatedTradeSummaryTableFilterSet,
    CumulativeNNMChartFilter,
    DistributionNNMChartFilter,
    CustomerAPIFilter,
    CustomerClaimFilter,
    CustomerClaimGroupByFilter,
    NegativeTermimalAccountPerProductFilterSet,
    ProfitAndLossPandasFilter,
)
from .fees import FeesAggregatedFilter, FeesFilter, FeesProductFilterSet
from .trades import (
    SubscriptionRedemptionFilterSet,
    SubscriptionRedemptionPortfolioFilterSet,
    TradeFilter,
    TradeInstrumentFilterSet,
    TradePortfolioFilter,
)



from datetime import timedelta

from django.db.models import Count, OuterRef, Subquery
from wbcore import filters as wb_filters
from wbcrm.models.accounts import Account
from wbfdm.models import Classification, Instrument

from wbportfolio.models import Portfolio, Product, Trade
from wbportfolio.models.transactions.claim import Claim

from .mixins import OppositeSharesFieldMethodMixin
from .utils import get_transaction_default_date_range


class TradeFilter(OppositeSharesFieldMethodMixin, wb_filters.FilterSet):
    transaction_date = wb_filters.DateRangeFilter(
        method=wb_filters.DateRangeFilter.base_date_range_filter_method,
        label="Date Range",
        initial=get_transaction_default_date_range,
    )

    portfolio = wb_filters.ModelChoiceFilter(
        label="Portfolio",
        queryset=Portfolio.objects.all(),
        endpoint=Portfolio.get_representation_endpoint(),
        value_key=Portfolio.get_representation_value_key(),
        label_key=Portfolio.get_representation_label_key(),
    )

    # we have to redefine the mixin fields because django_filters does not allow class extension with mixin
    opposite_shares = wb_filters.NumberFilter(
        field_name="shares",
        label="Opposite Shares",
        method="filter_opposite_shares",
        lookup_icon="±",
        lookup_label="Opposite",
    )
    opposite_approximate_shares = wb_filters.NumberFilter(
        field_name="shares",
        label="Opposite Approximite Shares",
        method="filter_opposite_approximate_shares",
        lookup_icon="≈±",
        lookup_label="Opposite Approximate (+- 10%)",
    )
    underlying_instrument = wb_filters.ModelChoiceFilter(
        label="Instrument",
        queryset=Instrument.objects.all(),
        endpoint=Instrument.get_representation_endpoint(),
        value_key=Instrument.get_representation_value_key(),
        label_key=Instrument.get_representation_label_key(),
        filter_params={"is_investable": True},
    )
    only_internal_trade = wb_filters.BooleanFilter(label="Only Internal Trade", method="boolean_only_internal_trade")

    completely_claimed = wb_filters.BooleanFilter(label="Completely Claimed", method="boolean_completely_claimed")
    completely_claimed_if_approved = wb_filters.BooleanFilter(
        label="Completely Claimed if approved", method="boolean_completely_claimed_if_approved"
    )

    claimed_shares__gte = wb_filters.NumberFilter(
        label="Claimed Shares", lookup_expr="gte", field_name="claimed_shares"
    )
    claimed_shares__lte = wb_filters.NumberFilter(
        label="Claimed Shares", field_name="claimed_shares", lookup_expr="lte"
    )
    claims = wb_filters.ModelChoiceFilter(
        label="Customer Accounts",
        queryset=Account.objects.all(),
        endpoint=Account.get_representation_endpoint(),
        value_key=Account.get_representation_value_key(),
        label_key=Account.get_representation_label_key(),
        method="filter_customer",
    )
    is_customer_trade = wb_filters.BooleanFilter(label="Is Customer Trade?", method="boolean_is_customer_trade")
    # register = wb_filters.ModelChoiceFilter(
    #     label="Register",
    #     queryset=Register.objects.all(),
    #     endpoint=Register.get_representation_endpoint(),
    #     value_key=Register.get_representation_value_key(),
    #     label_key = Register.get_representation_label_key(),
    # )
    marked_for_deletion = wb_filters.BooleanFilter(
        label="Marked For Deletion", initial=False, field_name="marked_for_deletion", lookup_expr="exact"
    )
    pivot_date = wb_filters.DateFilter(hidden=True, method="filter_pivot_date")

    def filter_pivot_date(self, queryset, name, value):
        if value:
            return queryset.filter(
                transaction_date__gte=value - timedelta(days=Trade.TRADE_WINDOW_INTERVAL),
                transaction_date__lte=value + timedelta(days=Trade.TRADE_WINDOW_INTERVAL),
            )
        return queryset

    def filter_customer(self, queryset, name, value):
        if value:
            accounts = value.get_descendants(include_self=True)
            queryset = queryset.annotate(
                filtered_claims=Subquery(
                    Claim.objects.filter(trade=OuterRef("pk"), account__in=accounts)
                    .values("account")
                    .annotate(c=Count("account"))
                    .values("c")[:1]
                )
            )
            return queryset.filter(filtered_claims__gt=0)
        return queryset

    def boolean_is_customer_trade(self, queryset, name, value):
        if value is True:
            return queryset.filter(transaction_subtype__in=[Trade.Type.SUBSCRIPTION, Trade.Type.REDEMPTION])
        elif value is False:
            return queryset.exclude(transaction_subtype__in=[Trade.Type.SUBSCRIPTION, Trade.Type.REDEMPTION])
        else:
            return queryset

    def boolean_completely_claimed(self, queryset, name, value):
        if value is True:
            return queryset.filter(completely_claimed=True)
        elif value is False:
            return queryset.filter(completely_claimed=False)
        else:
            return queryset

    def boolean_completely_claimed_if_approved(self, queryset, name, value):
        if value is True:
            return queryset.filter(completely_claimed_if_approved=True)
        elif value is False:
            return queryset.filter(completely_claimed_if_approved=False)
        else:
            return queryset

    def boolean_only_internal_trade(self, queryset, name, value):
        if value:
            return queryset.exclude(transaction_subtype__in=[Trade.Type.SUBSCRIPTION, Trade.Type.REDEMPTION]).filter(
                underlying_instrument__instrument_type__key="product"
            )
        return queryset

    class Meta:
        model = Trade
        fields = {
            "shares": ["gte", "lte", "exact"],
            "bank": ["exact", "icontains"],
            "price": ["gte", "lte", "exact"],
            "currency": ["exact"],
            "currency_fx_rate": ["gte", "exact", "lte"],
            "total_value": ["gte", "exact", "lte"],
            "total_value_fx_portfolio": ["gte", "exact", "lte"],
            "comment": ["icontains"],
            "portfolio": ["exact"],
            "register": ["exact"],
            # 'total_value_gross_fx_portfolio': ["gte", "exact", "lte"],
            "transaction_subtype": ["exact"],
            "pending": ["exact"],
        }


class TradePortfolioFilter(TradeFilter):
    portfolio = None

    class Meta:
        model = Trade
        fields = {
            "shares": ["gte", "lte", "exact"],
            "bank": ["exact", "icontains"],
            "price": ["gte", "lte", "exact"],
            "currency": ["exact"],
            "currency_fx_rate": ["gte", "exact", "lte"],
            "total_value": ["gte", "exact", "lte"],
            "total_value_fx_portfolio": ["gte", "exact", "lte"],
            "comment": ["icontains"],
            "register": ["exact"],
            "transaction_subtype": ["exact"],
        }


class SubscriptionRedemptionFilterSet(TradeFilter):
    is_customer_trade = None
    underlying_instrument = wb_filters.ModelChoiceFilter(
        label="Product",
        queryset=Product.objects.all(),
        endpoint=Product.get_representation_endpoint(),
        value_key=Product.get_representation_value_key(),
        label_key=Product.get_representation_label_key(),
    )
    underlying_instrument__classifications = wb_filters.ModelMultipleChoiceFilter(
        label="Classifications",
        queryset=Classification.objects.all(),
        endpoint=Classification.get_representation_endpoint(),
        value_key=Classification.get_representation_value_key(),
        label_key=Classification.get_representation_label_key(),
        filter_params={"instrument_type_key": "product"},
    )

    class Meta:
        model = Trade
        fields = {
            "shares": ["gte", "lte", "exact"],
            "bank": ["exact", "icontains"],
            "price": ["gte", "lte", "exact"],
            "currency": ["exact"],
            "total_value": ["gte", "exact", "lte"],
            "comment": ["icontains"],
            "portfolio": ["exact"],
            "register": ["exact"],
            # 'total_value_gross_fx_portfolio': ["gte", "exact", "lte"],
            "transaction_subtype": ["exact"],
            "marked_for_deletion": ["exact"],
            "custodian": ["exact"],
            "pending": ["exact"],
        }


class TradeInstrumentFilterSet(TradeFilter):
    is_customer_trade = underlying_instrument = completely_claimed = completely_claimed_if_approved = (
        claimed_shares__gte
    ) = claimed_shares__lte = claims = total_value_usd__gte = None

    class Meta:
        model = Trade
        fields = {
            "shares": ["gte", "lte", "exact"],
            "bank": ["exact", "icontains"],
            "price": ["gte", "lte", "exact"],
            "currency_fx_rate": ["gte", "exact", "lte"],
            "total_value": ["gte", "exact", "lte"],
            "total_value_fx_portfolio": ["gte", "exact", "lte"],
            "total_value_gross": ["gte", "exact", "lte"],
            "total_value_gross_fx_portfolio": ["gte", "exact", "lte"],
            "comment": ["icontains"],
            "portfolio": ["exact"],
            "register": ["exact"],
            "transaction_subtype": ["exact"],
        }


class SubscriptionRedemptionPortfolioFilterSet(TradeFilter):
    is_customer_trade = None

    class Meta:
        model = Trade
        fields = {
            "shares": ["gte", "lte", "exact"],
            "bank": ["exact", "icontains"],
            "price": ["gte", "lte", "exact"],
            "currency": ["exact"],
            "total_value": ["gte", "exact", "lte"],
            "comment": ["icontains"],
            "register": ["exact"],
            # 'total_value_gross_fx_portfolio': ["gte", "exact", "lte"],
            "transaction_subtype": ["exact"],
            "custodian": ["exact"],
        }



from django.urls import include, path
from wbcore.routers import WBCoreRouter

from wbportfolio import viewsets

router = WBCoreRouter()
router.register(r"assetposition", viewsets.AssetPositionModelViewSet, basename="assetposition")
router.register(r"adjustment", viewsets.AdjustmentModelViewSet, basename="adjustment")
router.register(r"product", viewsets.ProductModelViewSet, basename="product")
router.register(r"product_group", viewsets.ProductGroupModelViewSet, basename="product_group")
router.register(r"productcustomer", viewsets.ProductCustomerModelViewSet, basename="productcustomer")
router.register(r"productrepresentation", viewsets.ProductRepresentationViewSet, basename="productrepresentation")
router.register(
    r"product_grouprepresentation", viewsets.ProductGroupRepresentationViewSet, basename="product_grouprepresentation"
)
router.register(
    r"custodianrepresentation", viewsets.CustodianRepresentationViewSet, basename="custodianrepresentation"
)


router.register(r"portfolio", viewsets.PortfolioModelViewSet, basename="portfolio")
router.register(r"portfoliocashflow", viewsets.DailyPortfolioCashFlowModelViewSet, basename="portfoliocashflow")
router.register(r"portfolioswingpricing", viewsets.PortfolioSwingPricingModelViewSet, basename="portfolioswingpricing")
router.register(r"portfoliocashtarget", viewsets.PortfolioCashTargetModelViewSet, basename="portfoliocashtarget")

router.register(
    r"portfoliorepresentation", viewsets.PortfolioRepresentationViewSet, basename="portfoliorepresentation"
)

router.register(r"portfoliorole", viewsets.PortfolioRoleModelViewSet, basename="portfoliorole")


router.register(r"fees", viewsets.FeesModelViewSet, basename="fees")
router.register(r"trade", viewsets.TradeModelViewSet, basename="trade")
router.register(
    r"subscriptionredemption", viewsets.SubscriptionRedemptionModelViewSet, basename="subscriptionredemption"
)

router.register(r"traderepresentation", viewsets.TradeRepresentationViewSet, basename="traderepresentation")
router.register(
    r"orderproposalrepresentation", viewsets.OrderProposalRepresentationViewSet, basename="orderproposalrepresentation"
)
router.register(
    r"rebalancingmodelrepresentation",
    viewsets.RebalancingModelRepresentationViewSet,
    basename="rebalancingmodelrepresentation",
)
router.register(
    r"rebalancerrepresentation",
    viewsets.RebalancerRepresentationViewSet,
    basename="rebalancerrepresentation",
)
router.register(
    r"rebalancer",
    viewsets.RebalancerModelViewSet,
    basename="rebalancer",
)
router.register(
    r"feesproductperformance", viewsets.ProductPerformanceFeesModelViewSet, basename="feesproductperformance"
)
router.register(r"aumchart", viewsets.InstrumentPriceAUMDataChartView, basename="aumchart")

router.register(r"custodian", viewsets.CustodianModelViewSet, basename="custodian")

router.register(r"orderproposal", viewsets.OrderProposalModelViewSet, basename="orderproposal")

router.register(
    r"assetandnetnewmoneyprogression",
    viewsets.AssetAndNetNewMoneyProgressionChartViewSet,
    basename="assetandnetnewmoneyprogression",
)


# Subrouter for Portfolio
portfolio_router = WBCoreRouter()
portfolio_router.register(
    r"portfoliocashflow", viewsets.DailyPortfolioCashFlowModelViewSet, basename="portfolio-portfoliocashflow"
)
portfolio_router.register(r"asset", viewsets.AssetPositionPortfolioModelViewSet, basename="portfolio-asset")
portfolio_router.register(r"contributor", viewsets.ContributorPortfolioChartView, basename="portfolio-contributor")
portfolio_router.register(r"trade", viewsets.TradePortfolioModelViewSet, basename="portfolio-trade")
portfolio_router.register(
    r"instrument", viewsets.InstrumentPortfolioThroughPortfolioModelViewSet, basename="portfolio-instrument"
)
portfolio_router.register(
    r"preferredclassification",
    viewsets.InstrumentPreferedClassificationThroughProductModelViewSet,
    basename="portfolio-preferredclassification",
)
portfolio_router.register(
    r"orderproposal", viewsets.OrderProposalPortfolioModelViewSet, basename="portfolio-orderproposal"
)
portfolio_router.register(
    r"dependencyportfolio", viewsets.PortfolioPortfolioThroughModelViewSet, basename="portfolio-dependencyportfolio"
)
portfolio_router.register(
    r"modelcompositionpandas",
    viewsets.CompositionModelPortfolioPandasView,
    basename="portfolio-modelcompositionpandas",
)

portfolio_router.register(
    r"distributionchart",
    viewsets.DistributionChartViewSet,
    basename="portfolio-distributionchart",
)
portfolio_router.register(
    r"distributiontable",
    viewsets.DistributionTableViewSet,
    basename="portfolio-distributiontable",
)
portfolio_router.register(
    r"esgaggregation",
    viewsets.ESGMetricAggregationPortfolioPandasViewSet,
    basename="portfolio-esgaggregation",
)
portfolio_router.register(
    r"treegraphchart",
    viewsets.PortfolioTreeGraphChartViewSet,
    basename="portfolio-treegraphchart",
)
portfolio_router.register(
    r"topdowncomposition",
    viewsets.TopDownPortfolioCompositionPandasAPIView,
    basename="portfolio-topdowncomposition",
)

router.register(r"register", viewsets.RegisterModelViewSet, basename="register")
router.register(r"registerrepresentation", viewsets.RegisterRepresentationViewSet, basename="registerrepresentation")


# Subrouter for Products
product_router = WBCoreRouter()
product_router.register(r"nominalchart", viewsets.NominalProductChartView, basename="product-nominalchart")
product_router.register(r"aumchart", viewsets.AUMProductChartView, basename="product-aumchart")
product_router.register(r"claim", viewsets.ClaimProductModelViewSet, basename="product-claim")
product_router.register(r"feesaggregated", viewsets.FeesAggregatedProductPandasView, basename="product-feesaggregated")
product_router.register(r"fees", viewsets.FeesProductModelViewSet, basename="product-fees")

# Subrouter for Order Proposal
order_proposal_router = WBCoreRouter()
order_proposal_router.register(r"trade", viewsets.OrderOrderProposalModelViewSet, basename="orderproposal-order")

trade_router = WBCoreRouter()
trade_router.register(r"claim", viewsets.ClaimTradeModelViewSet, basename="trade-claim")

instrument_router = WBCoreRouter()

instrument_router.register(r"trade", viewsets.TradeInstrumentModelViewSet, basename="instrument-trade")
instrument_router.register(
    r"subscriptionredemption",
    viewsets.SubscriptionRedemptionInstrumentModelViewSet,
    basename="instrument-subscriptionredemption",
)
instrument_router.register(
    r"custodiandistribution",
    viewsets.CustodianDistributionInstrumentChartViewSet,
    basename="instrument-custodiandistribution",
)
instrument_router.register(
    r"customerdistribution",
    viewsets.CustomerDistributionInstrumentChartViewSet,
    basename="instrument-customerdistribution",
)
instrument_router.register(r"asset", viewsets.AssetPositionInstrumentModelViewSet, basename="instrument-asset")

instrument_router.register(
    r"portfoliorole", viewsets.PortfolioRoleInstrumentModelViewSet, basename="instrument-portfoliorole"
)

instrument_router.register(r"adjustment", viewsets.AdjustmentEquityModelViewSet, basename="instrument-adjustment")
instrument_router.register(
    r"assetpositionchart",
    viewsets.AssetPositionUnderlyingInstrumentChartViewSet,
    basename="instrument-assetpositionchart",
)

# app_name = 'wbportfolio'

router.register(r"productperformance", viewsets.PerformancePandasView, basename="productperformance")
router.register(
    r"productperformancennmlist",
    viewsets.ProductPerformanceNetNewMoneyListViewSet,
    basename="productperformancennmlist",
)
router.register(
    r"productperformancecomparison", viewsets.PerformanceComparisonPandasView, basename="productperformancecomparison"
)
router.register(r"productcashposition", viewsets.CashPositionPortfolioPandasAPIView, basename="productcashposition")
router.register(r"assetpositiongroupby", viewsets.AssetPositionPandasView, basename="assetpositiongroupby")
router.register(
    r"aggregatedassetpositionliquidity",
    viewsets.AggregatedAssetPositionLiquidityPandasView,
    basename="aggregatedassetpositionliquidity",
)

router.register(r"accountreconciliation", viewsets.AccountReconciliationModelViewSet, basename="accountreconciliation")
router.register(
    r"accountreconciliationline", viewsets.AccountReconciliationLineModelViewSet, basename="accountreconciliationline"
)

account_reconciliation_router = WBCoreRouter()
account_reconciliation_router.register(
    r"accountreconciliationline",
    viewsets.AccountReconciliationLineModelViewSet,
    basename="accountreconciliation-accountreconciliationline",
)

# Claim routers
router.register(r"claimrepresentation", viewsets.ClaimRepresentationViewSet, basename="claimrepresentation")
router.register(r"claim", viewsets.ClaimModelViewSet, basename="claim")
router.register(r"claim-api", viewsets.ClaimAPIModelViewSet, basename="claim-api")
router.register(
    r"negativeaccountproduct",
    viewsets.NegativeTermimalAccountPerProductModelViewSet,
    basename="negativeaccountproduct",
)
router.register(
    r"aumtable",
    viewsets.ConsolidatedTradeSummaryTableView,
    basename="aumtable",
)
router.register(
    r"consolidatedtradesummarydistributionchart",
    viewsets.ConsolidatedTradeSummaryDistributionChart,
    basename="consolidatedtradesummarydistributionchart",
)
router.register(
    r"cumulativennmchart",
    viewsets.CumulativeNNMChartView,
    basename="cumulativennmchart",
)

router.register(
    r"pnltable",
    viewsets.ProfitAndLossPandasView,
    basename="pnltable",
)

account_router = WBCoreRouter()
account_router.register(r"claim", viewsets.ClaimAccountModelViewSet, basename="account-claim")


entry_router = WBCoreRouter()
entry_router.register(r"claim", viewsets.ClaimEntryModelViewSet, basename="entry-claim")


urlpatterns = [
    path("", include(router.urls)),
    path("product/<int:product_id>/", include(product_router.urls)),
    path("trade/<int:trade_id>/", include(trade_router.urls)),
    path("portfolio/<int:portfolio_id>/", include(portfolio_router.urls)),
    path("orderproposal/<int:order_proposal_id>/", include(order_proposal_router.urls)),
    path("instrument/<int:instrument_id>/", include(instrument_router.urls)),
    path("account/<int:account_id>/", include(account_router.urls)),
    path("entry/<int:entry_id>/", include(entry_router.urls)),
    path("accountreconciliation/<int:accountreconciliation_id>/", include(account_reconciliation_router.urls)),
]



# Generated by Django 4.2.6 on 2023-10-31 14:51

from decimal import Decimal

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("wbportfolio", "0048_alter_trade_status"),
    ]

    operations = [
        migrations.AddField(
            model_name="trade",
            name="claimed_shares",
            field=models.DecimalField(
                decimal_places=4,
                default=Decimal("0"),
                help_text="The number of shares that were claimed.",
                max_digits=15,
                verbose_name="Claimed Shares",
            ),
        ),
    ]



from contextlib import suppress

from wbcore.contrib.icons import WBIcon
from wbcore.metadata.configs import buttons as bt
from wbfdm.models.instruments import Instrument


class InstrumentButtonMixin:
    @classmethod
    def add_instrument_request_button(self, request=None, view=None, pk=None, **kwargs):
        buttons = [
            bt.WidgetButton(key="assets", label="Implemented Portfolios (Assets)"),
            # bt.WidgetButton(
            #     key="adjustments",
            #     label="Adjustments",
            #     icon=WBIcon.DATA_LIST.icon,
            # ),
        ]
        with suppress(Instrument.DoesNotExist):
            instrument = Instrument.objects.get(id=pk)
            asset_instrument_btn_label = "Asset Portfolio"
            if instrument.portfolio:
                buttons.extend(
                    [
                        bt.WidgetButton(key="portfolio_positions", label=asset_instrument_btn_label),
                        bt.DropDownButton(
                            label="Charts",
                            icon=WBIcon.UNFOLD.icon,
                            buttons=(
                                bt.WidgetButton(
                                    key="portfolio_positions_contributors", label="Contributors (Computed)"
                                ),
                                bt.WidgetButton(key="distribution_chart", label="Distribution Chart"),
                                bt.WidgetButton(key="distribution_table", label="Distribution Table"),
                                bt.WidgetButton(key="assetschart", label="Portfolio Allocation"),
                            ),
                        ),
                    ]
                )
        return bt.DropDownButton(
            label="Portfolio",
            icon=WBIcon.UNFOLD.icon,
            buttons=buttons,
        )

    @classmethod
    def add_transactions_request_button(self, request=None, view=None, pk=None, **kwargs):
        return bt.DropDownButton(
            label="Transactions",
            icon=WBIcon.UNFOLD.icon,
            buttons=(
                bt.WidgetButton(key="portfolio_transactions", label="Transactions"),
                bt.WidgetButton(key="portfolio_trades", label="Trades"),
                bt.WidgetButton(key="instrument_subscriptionsredemptions", label="Subscriptions/Redemptions"),
                bt.WidgetButton(key="instrument_trades", label="Trades (Implemented)"),
                bt.WidgetButton(key="product_fees", label="Fees"),
                bt.WidgetButton(key="product_aggregatedfees", label="Aggregated Fees"),
                bt.DropDownButton(
                    label="Charts",
                    icon=WBIcon.UNFOLD.icon,
                    buttons=(
                        bt.WidgetButton(key="tradechart", label="Nominal"),
                        bt.WidgetButton(key="aumchart", label="AUM"),
                        bt.WidgetButton(key="custodiandistribution", label="Custodian Distribution"),
                        bt.WidgetButton(key="customerdistribution", label="Customer Distribution"),
                    ),
                ),
            ),
        )



import factory
from faker import Faker
from pandas._libs.tslibs.offsets import BDay

from wbportfolio.models import OrderProposal

fake = Faker()


class OrderProposalFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = OrderProposal

    trade_date = factory.LazyAttribute(lambda o: (fake.date_object() + BDay(1)).date())
    comment = factory.Faker("paragraph")
    portfolio = factory.SubFactory("wbportfolio.factories.PortfolioFactory")
    creator = factory.SubFactory("wbcore.contrib.directory.factories.PersonFactory")



# Generated by Django 4.1.9 on 2023-07-13 12:00

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [("wbportfolio", "0046_add_product_default_account")]

    operations = [
        migrations.AddIndex(
            model_name="fees",
            index=models.Index(fields=["linked_product"], name="wbportfolio_linked__a49c5d_idx"),
        ),
    ]



import random
from decimal import Decimal

import factory
from faker import Faker

from wbportfolio.models import Order

fake = Faker()


class OrderFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = Order

    order_proposal = factory.SubFactory("wbportfolio.factories.OrderProposalFactory")
    currency_fx_rate = Decimal(1.0)
    fees = Decimal(0.0)
    underlying_instrument = factory.SubFactory("wbfdm.factories.InstrumentFactory")
    shares = factory.Faker("pydecimal", min_value=10, max_value=1000, right_digits=4)
    price = factory.LazyAttribute(lambda o: random.randint(10, 10000))



# Generated by Django 4.1.10 on 2023-07-26 08:52

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("wbfdm", "0005_instrument_corporate_action_backend"),
        ("wbportfolio", "0044_alter_instrumentlist_identifier"),
    ]

    operations = [
        migrations.AlterField(
            model_name="instrument",
            name="financial_instrument",
            field=models.OneToOneField(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="pms_instrument",
                to="wbfdm.instrument",
                verbose_name="Financial Instrument",
            ),
        ),
    ]



from .accounts import *
from .controversy_portfolio import *
from .exposure_portfolio import *
from .instrument_list_portfolio import *
from .liquidity_stress_instrument import *
from .stop_loss_instrument import *
from .stop_loss_portfolio import *
from .ucits_portfolio import *
from .accounts import *
from .product_integrity import *
from .liquidity_risk import *



from wbcore.metadata.configs.buttons.enums import Button
from wbcore.metadata.configs.buttons.view_config import ButtonViewConfig


class OrderOrderProposalButtonConfig(ButtonViewConfig):
    def get_create_buttons(self):
        return {
            Button.SAVE_AND_CLOSE.value,
        }



from typing import Optional

from django.shortcuts import get_object_or_404
from django.utils.translation import gettext_lazy as _
from wbcore.contrib.color.enums import WBColor
from wbcore.enums import Unit
from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display.instance_display.shortcuts import (
    Display,
    create_simple_display,
)
from wbcore.metadata.configs.display.instance_display.utils import repeat_field
from wbcore.metadata.configs.display.view_config import DisplayViewConfig

from wbportfolio.models import Order, OrderProposal

ORDER_STATUS_LEGENDS = dp.Legend(
    key="has_warnings",
    items=[
        dp.LegendItem(icon=WBColor.YELLOW_DARK.value, label=_("Warning"), value=True),
    ],
)

ORDER_STATUS_FORMATTING = dp.Formatting(
    column="has_warnings",
    formatting_rules=[
        dp.FormattingRule(
            style={"backgroundColor": WBColor.YELLOW_DARK.value},
            condition=("==", True),
        )
    ],
)
ORDER_TYPE_FORMATTING_RULES = [
    dp.FormattingRule(
        style={"color": WBColor.RED_DARK.value, "fontWeight": "bold"},
        condition=("==", Order.Type.SELL.name),
    ),
    dp.FormattingRule(
        style={"color": WBColor.RED.value, "fontWeight": "bold"},
        condition=("==", Order.Type.DECREASE.name),
    ),
    dp.FormattingRule(
        style={"color": WBColor.GREEN.value, "fontWeight": "bold"},
        condition=("==", Order.Type.INCREASE.name),
    ),
    dp.FormattingRule(
        style={"color": WBColor.GREEN_DARK.value, "fontWeight": "bold"},
        condition=("==", Order.Type.BUY.name),
    ),
    dp.FormattingRule(
        style={"color": WBColor.GREY.value, "fontWeight": "bold"},
        condition=("==", Order.Type.NO_CHANGE.name),
    ),
]

VALUE_FORMATTING_RULES = [
    dp.FormattingRule(
        style={"color": WBColor.RED_DARK.value, "fontWeight": "bold"},
        condition=("<", 0),
    ),
    dp.FormattingRule(
        style={"color": WBColor.GREEN_DARK.value, "fontWeight": "bold"},
        condition=(">", 0),
    ),
]


class OrderOrderProposalDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        order_proposal = get_object_or_404(OrderProposal, pk=self.view.kwargs.get("order_proposal_id", None))
        fields = [
            dp.Field(
                label="Instrument",
                open_by_default=True,
                key=None,
                children=[
                    dp.Field(key="underlying_instrument", label="Name", width=Unit.PIXEL(250)),
                    dp.Field(key="underlying_instrument_isin", label="ISIN", width=Unit.PIXEL(125)),
                    dp.Field(key="underlying_instrument_ticker", label="Ticker", width=Unit.PIXEL(100)),
                    dp.Field(
                        key="underlying_instrument_refinitiv_identifier_code", label="RIC", width=Unit.PIXEL(100)
                    ),
                    dp.Field(key="underlying_instrument_instrument_type", label="Asset Class", width=Unit.PIXEL(125)),
                ],
            ),
            dp.Field(
                label="Weight",
                open_by_default=False,
                key=None,
                children=[
                    dp.Field(key="effective_weight", label="Effective Weight", show="open", width=Unit.PIXEL(150)),
                    dp.Field(key="target_weight", label="Target Weight", show="open", width=Unit.PIXEL(150)),
                    dp.Field(
                        key="weighting",
                        label="Delta Weight",
                        formatting_rules=VALUE_FORMATTING_RULES,
                        width=Unit.PIXEL(150),
                    ),
                ],
            ),
        ]
        if not order_proposal.portfolio.only_weighting:
            fields.append(
                dp.Field(
                    label="Shares",
                    open_by_default=False,
                    key=None,
                    children=[
                        dp.Field(key="effective_shares", label="Effective Shares", show="open", width=Unit.PIXEL(150)),
                        dp.Field(key="target_shares", label="Target Shares", show="open", width=Unit.PIXEL(150)),
                        dp.Field(
                            key="shares",
                            label="Shares",
                            formatting_rules=VALUE_FORMATTING_RULES,
                            width=Unit.PIXEL(150),
                        ),
                    ],
                )
            )
            fields.append(
                dp.Field(
                    label="Total Value",
                    open_by_default=False,
                    key=None,
                    children=[
                        dp.Field(
                            key="effective_total_value_fx_portfolio",
                            label="Effective Total Value",
                            show="open",
                            width=Unit.PIXEL(150),
                        ),
                        dp.Field(
                            key="target_total_value_fx_portfolio",
                            label="Target Total Value",
                            show="open",
                            width=Unit.PIXEL(150),
                        ),
                        dp.Field(
                            key="total_value_fx_portfolio",
                            label="Total Value",
                            formatting_rules=VALUE_FORMATTING_RULES,
                            width=Unit.PIXEL(150),
                        ),
                    ],
                )
            )
        fields.append(
            dp.Field(
                label="Information",
                open_by_default=False,
                key=None,
                children=[
                    dp.Field(
                        key="order_type",
                        label="Direction",
                        formatting_rules=ORDER_TYPE_FORMATTING_RULES,
                        width=Unit.PIXEL(125),
                    ),
                    dp.Field(key="comment", label="Comment", width=Unit.PIXEL(250)),
                    dp.Field(key="order", label="Order", show="open", width=Unit.PIXEL(100)),
                ],
            )
        )
        return dp.ListDisplay(
            fields=fields,
            legends=[ORDER_STATUS_LEGENDS],
            formatting=[ORDER_STATUS_FORMATTING],
        )

    def get_instance_display(self) -> Display:
        order_proposal = get_object_or_404(OrderProposal, pk=self.view.kwargs.get("order_proposal_id", None))

        fields = [
            ["company", "security", "underlying_instrument"],
            ["effective_weight", "target_weight", "weighting"],
        ]
        if not order_proposal.portfolio.only_weighting:
            fields.append(["effective_shares", "target_shares", "shares"])
        fields.append([repeat_field(3, "comment")])
        return create_simple_display(fields)



from wbcore.menus import ItemPermission, MenuItem

ACCOUNT_RECONCILIATION_MENU_ITEM = MenuItem(
    label="Account Reconciliations",
    endpoint="wbportfolio:accountreconciliation-list",
    permission=ItemPermission(permissions=["wbportfolio.view_accountreconciliation"]),
)



from typing import Optional

from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display import Display, create_simple_display
from wbcore.metadata.configs.display.view_config import DisplayViewConfig


class RebalancerDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="portfolio", label="Portfolio"),
                dp.Field(key="rebalancing_model", label="Rebalancing Model"),
                dp.Field(key="approve_order_proposal_automatically", label="Approve automatically"),
                dp.Field(key="frequency_repr", label="Frequency"),
                dp.Field(key="activation_date", label="Activation Date"),
            ],
        )

    def get_instance_display(self) -> Display:
        return create_simple_display(
            [
                ["rebalancing_model", "approve_order_proposal_automatically"],
                ["frequency", "activation_date"],
                ["rebalancing_dates", "rebalancing_dates"],
            ]
        )



from django.apps import apps
from django.contrib.contenttypes.models import ContentType
from django.db.models import Q
from django.dispatch import receiver
from rest_framework.reverse import reverse
from wbcore.contrib.icons import WBIcon
from wbcore.metadata.configs import buttons as bt
from wbcore.signals.instance_buttons import add_instance_button

from wbportfolio.viewsets.products import ProductModelViewSet


@receiver(add_instance_button, sender=ProductModelViewSet)
def add_report_buttons(sender, many, *args, request=None, **kwargs):
    report_buttons = []
    try:
        if instrument_id := kwargs.get("pk", None):
            content_type_product = ContentType.objects.get(app_label="wbportfolio", model="product")
            content_type_instrument = ContentType.objects.get(app_label="wbfdm", model="instrument")
            Report = apps.get_model("wbreport", "Report")
            for report in Report.objects.filter(object_id=instrument_id, is_active=True).filter(
                Q(content_type=content_type_product) | Q(content_type=content_type_instrument)
            ):
                if primary_version := report.primary_version:
                    report_buttons.append(
                        bt.HyperlinkButton(
                            label=f"{report.title} Report",
                            endpoint=reverse(
                                "public_report:report_version", args=[primary_version.lookup], request=request
                            ),
                        ),
                    )
                    if request.user.profile.is_internal or request.user.is_superuser:
                        report_buttons.append(
                            bt.HyperlinkButton(
                                label=f"{report.title} Widget",
                                endpoint=reverse("wbreport:report-detail", args=[report.id], request=request),
                            ),
                        )
    except LookupError:
        pass
    if report_buttons:
        return bt.DropDownButton(label="Reports", buttons=tuple(report_buttons), icon=WBIcon.NOTEBOOK.icon)



from django.apps import AppConfig
from django.apps import apps as global_apps
from django.db import DEFAULT_DB_ALIAS
from django.db.models.signals import post_migrate
from django.utils.module_loading import autodiscover_modules


class WbportfolioConfig(AppConfig):
    name = "wbportfolio"

    def ready(self):
        def autodiscover_backends(
            app_config, verbosity=2, interactive=True, using=DEFAULT_DB_ALIAS, apps=global_apps, **kwargs
        ):
            # we wrap the autodiscover into a post_migrate receiver because we expect db calls
            autodiscover_modules("rebalancing")

        post_migrate.connect(
            autodiscover_backends,
            dispatch_uid="wbportfolio.autodiscover_rebalancing",
        )



from rest_framework.reverse import reverse
from wbcore import serializers
from wbcore.contrib.icons import WBIcon
from wbcore.enums import RequestType
from wbcore.metadata.configs import buttons as bt
from wbcore.metadata.configs.buttons.view_config import ButtonViewConfig
from wbcore.metadata.configs.display.instance_display.shortcuts import (
    create_simple_display,
)
from wbcore.utils.urls import get_urlencode_endpoint
from wbcrm.models.accounts import Account
from wbcrm.serializers.accounts import TerminalAccountRepresentationSerializer


class TransferTradeSerializer(serializers.Serializer):
    transfer_date = serializers.DateField(label="Transfer Date", required=True)
    from_account = serializers.PrimaryKeyRelatedField(
        queryset=Account.objects.all(), label="From Account", required=True
    )
    _from_account = TerminalAccountRepresentationSerializer(source="from_account")

    to_account = serializers.PrimaryKeyRelatedField(queryset=Account.objects.all(), label="To Account", required=True)
    _to_account = TerminalAccountRepresentationSerializer(source="to_account")


class QuickClaimSerializer(serializers.Serializer):
    account = serializers.PrimaryKeyRelatedField(queryset=Account.objects.all(), label="Account", required=True)
    _account = TerminalAccountRepresentationSerializer(source="account")


class ConsolidatedTradeSummaryButtonConfig(ButtonViewConfig):
    def get_custom_buttons(self) -> set:
        return {
            bt.WidgetButton(
                endpoint=get_urlencode_endpoint(
                    reverse(
                        "wbportfolio:consolidatedtradesummarydistributionchart-list", args=[], request=self.request
                    ),
                    self.request.GET,
                ),
                label="Distribution NNM Chart",
            ),
            bt.WidgetButton(
                endpoint=get_urlencode_endpoint(
                    reverse("wbportfolio:cumulativennmchart-list", args=[], request=self.request), self.request.GET
                ),
                label="Cumulative NNM Chart",
            ),
        }


class ClaimTradeButtonConfig(ButtonViewConfig):
    def get_custom_buttons(self):
        return {
            bt.ActionButton(
                method=RequestType.POST,
                action_label="Trade transferred.",
                endpoint=reverse(
                    "wbportfolio:trade-claim-transfer-trade", request=self.request, args=[self.view.kwargs["trade_id"]]
                ),
                description_fields="Do you want to transfer this trade?",
                label="Transfer Trade",
                icon=WBIcon.DEAL.icon,
                confirm_config=bt.ButtonConfig(label="Transfer"),
                cancel_config=bt.ButtonConfig(label="Cancel"),
                serializer=TransferTradeSerializer,
                instance_display=create_simple_display([["transfer_date"], ["from_account"], ["to_account"]]),
            ),
            bt.ActionButton(
                method=RequestType.POST,
                action_label="Trade claimed.",
                endpoint=reverse(
                    "wbportfolio:trade-claim-quick-claim", request=self.request, args=[self.view.kwargs["trade_id"]]
                ),
                description_fields="Do you want to quick claim this trade?",
                label="Quick Claim",
                icon=WBIcon.EURO.icon,
                confirm_config=bt.ButtonConfig(label="Claim"),
                cancel_config=bt.ButtonConfig(label="Cancel"),
                serializer=QuickClaimSerializer,
                instance_display=create_simple_display([["account"]]),
            ),
        }



from .adjustments import AdjustmentDisplayConfig
from .rebalancing import RebalancerDisplayConfig
from .assets import (
    AssetPositionDisplayConfig,
    AssetPositionInstrumentDisplayConfig,
    AssetPositionPortfolioDisplayConfig,
    CashPositionPortfolioDisplayConfig,
    CompositionModelPortfolioPandasDisplayConfig,
    DistributionTableDisplayConfig,
)

from .claim import (
    ConsolidatedTradeSummaryDisplayConfig,
    ClaimDisplayConfig,
    NegativeTermimalAccountPerProductDisplayConfig,
    ProfitAndLossPandasDisplayConfig,
)

from .custodians import CustodianDisplayConfig

from .fees import (
    FeesAggregatedProductPandasDisplayConfig,
    FeesDisplayConfig,
    FeesProductDisplayConfig,
)
from .portfolios import (
    PortfolioDisplayConfig,
    PortfolioPortfolioThroughModelDisplayConfig,
    TopDownPortfolioCompositionPandasDisplayConfig
)
from .positions import (
    AggregatedAssetPositionLiquidityDisplayConfig,
    AssetPositionPandasDisplayConfig,
)
from .product_groups import ProductGroupDisplayConfig
from .product_performance import (
    PerformanceComparisonDisplayConfig,
    PerformancePandasDisplayConfig,
    ProductPerformanceNetNewMoneyDisplayConfig,
)
from .products import (
    ProductCustomerDisplayConfig,
    ProductDisplayConfig,
    ProductPerformanceFeesDisplayConfig,
)
from .registers import RegisterDisplayConfig
from .roles import PortfolioRoleDisplayConfig, PortfolioRoleInstrumentDisplayConfig
from .trades import (
    SubscriptionRedemptionDisplayConfig,
    TradeDisplayConfig,
    TradePortfolioDisplayConfig,
)
from .portfolio_relationship import (
    PortfolioInstrumentPreferredClassificationThroughDisplayConfig,
    InstrumentPortfolioThroughPortfolioModelDisplayConfig,
)
from .portfolio_cash_flow import DailyPortfolioCashFlowDisplayConfig
from .esg import ESGMetricAggregationPortfolioPandasDisplayConfig
from .reconciliations import AccountReconciliationDisplayViewConfig, AccountReconciliationLineDisplayViewConfig



from wbcore.menus import ItemPermission, MenuItem

from wbportfolio.permissions import is_manager

GAIN_MENUITEM = MenuItem(
    label="Product Fees",
    endpoint="wbportfolio:feesproductperformance-list",
    permission=ItemPermission(permissions=["wbportfolio.view_product", "wbportfolio.view_fees"], method=is_manager),
)
FEES_MENUITEM = MenuItem(
    label="Fees",
    endpoint="wbportfolio:fees-list",
    permission=ItemPermission(permissions=["wbportfolio.view_fees"], method=is_manager),
)
